// Copyright 6/4/2019 Varjo Technologies Oy. All Rights Reserved.

#include "VarjoEyeTrackerFunctionLibrary.h"
#include "VarjoGaze.h"
#include "Engine/Engine.h"
#include "VarjoEyeTrackerModule.h"
#include "../../VarjoHMD/Private/VarjoHMD.h"
#include "Templates/SharedPointer.h"

FVarjoEyeTracker* GetVarjoEyeTracker()
{
	if (GEngine == nullptr)
	{
		return nullptr;
	}

	if (GEngine->XRSystem.IsValid() && (GEngine->XRSystem->GetSystemName() == FName("VarjoHMD"))) //works only if VarjoHMD is connected
	{
		if (GEngine->EyeTrackingDevice)
		{
			return StaticCastSharedPtr<FVarjoEyeTracker>(GEngine->EyeTrackingDevice).Get();
		}
	}
	return nullptr;
}

UVarjoEyeTrackerFunctionLibrary::UVarjoEyeTrackerFunctionLibrary(const FObjectInitializer& ObjectInitializer)
: Super(ObjectInitializer)
{
}

EVarjoEyeTrackingCalibrationStatus UVarjoEyeTrackerFunctionLibrary::GetCalibrationStatus()
{
	FVarjoEyeTracker* eyeTracker = GetVarjoEyeTracker();
	if (eyeTracker)
	{
		return eyeTracker->GetCalibrationStatus();
	}

	return EVarjoEyeTrackingCalibrationStatus::None;
}

bool UVarjoEyeTrackerFunctionLibrary::RequestGazeCalibration()
{
	FVarjoEyeTracker* eyeTracker = GetVarjoEyeTracker();
	if (eyeTracker)
	{
		eyeTracker->RequestGazeCalibration();
		return true;
	}
	return false;
}

void UVarjoEyeTrackerFunctionLibrary::GetGazeRay(FVector& start, FVector& direction, float& focusDistance)
{
	FVarjoEyeTracker* eyeTracker = GetVarjoEyeTracker();
	if (eyeTracker)
	{
		eyeTracker->GetGazeRay(start, direction, focusDistance);
	}
}

bool UVarjoEyeTrackerFunctionLibrary::GetEyeTrackerGazeData(FVarjoEyeTrackingData& eyeTrackingData)
{
	FVarjoEyeTracker* eyeTracker = GetVarjoEyeTracker();
	if (eyeTracker)
	{
		eyeTracker->GetVarjoEyeTrackerGazeData(eyeTrackingData);
		return true;
	}
	return false;
}

FString Double3ToString(double data[3])
{
	return FString::Printf(TEXT("%f, %f, %f"), data[0], data[1], data[2]);
}

TArray<FString> UVarjoEyeTrackerFunctionLibrary::GetGazeStringArray()
{
	FVarjoEyeTracker* eyeTracker = GetVarjoEyeTracker();
	TArray<varjo_Gaze> gazeData;
	if (eyeTracker)
	{
		eyeTracker->GetGazeRawArray(gazeData);
	}
	
	TArray<FString> result;
	FString header(TEXT("Frame, TargetId, LeftEyeGaze, RightEyeGaze, CombinedGaze, LeftEyePos, RightEyePos, CombinedPos, TargetPos, TargetDist, LeftEyeDist, RightEyeDist, CombinedDist"));
	for (auto& gaze : gazeData)
	{
		FString strHeader = FString::Printf(TEXT("%lld, %d"), gaze.frameNumber, -1);
		FString strGaze = FString::Printf(TEXT("%s, %s, %s"), *Double3ToString(gaze.leftEye.forward), *Double3ToString(gaze.rightEye.forward), *Double3ToString(gaze.gaze.forward));
		FString strPos = FString::Printf(TEXT("%s, %s, %s"), *Double3ToString(gaze.leftEye.origin), *Double3ToString(gaze.rightEye.origin), *Double3ToString(gaze.gaze.origin));
		result.Add(strHeader + ", " + strGaze + ", " + strPos);
	}

	return result;
}

TArray<FVarjoEyeTrackingData> UVarjoEyeTrackerFunctionLibrary::GetGazeArray()
{
	FVarjoEyeTracker* eyeTracker = GetVarjoEyeTracker();
	TArray<FVarjoEyeTrackingData> result;
	if (eyeTracker)
	{
		eyeTracker->GetGazeArray(result);
	}
	return std::move(result);
}

#include "Misc/FileHelper.h"
#include "HAL/PlatformFilemanager.h"

FString Vec3ToCSV(const FVector& vec)
{
	return FString::Printf(TEXT("%f, %f, %f"), vec.X, vec.Y, vec.Z);
}

float RayToPointDistance(FVector point, FVector rayOrigin, FVector rayDirection)
{
	FVector u = (point - rayOrigin);
	FVector v = rayDirection;
	return (FVector::Distance(u, FVector::DotProduct(u, v) * v));
}

bool UVarjoEyeTrackerFunctionLibrary::SaveGazeData(FString SaveDirectory, FString FileName, const TArray<FVarjoEyeTrackingData>& gazeArray)
{
	FString SavePath = SaveDirectory + FileName;

	FString dataString(TEXT("Frame;TargetId;LeftEyeGaze;RightEyeGaze;CombinedGaze;LeftEyePos;RightEyePos;CombinedPos;TargetPos;TargetDist;LeftEyeDist;RightEyeDist;CombinedDist"));
	dataString += LINE_TERMINATOR;
	for (auto& gaze : gazeArray)
	{
		if (gaze.rawData.frameNumber <= 0)
		{
			continue;
		}
		
		dataString += FString::Printf(TEXT("%lld;%d;"), gaze.rawData.frameNumber, gaze.targetId);
		if (gaze.status == varjo_GazeStatus_Invalid)
		{
			dataString += "Invalid;Invalid;Invalid;";
		}
		else
		{
			dataString += Vec3ToCSV(gaze.leftEye.dir) + ";" + Vec3ToCSV(gaze.rightEye.dir) + ";" + *Vec3ToCSV(gaze.gaze.dir) + ";";
		}

		if (gaze.status == varjo_GazeStatus_Invalid)
		{
			dataString += "Invalid;Invalid;Invalid;";
		}
		else
		{
			dataString += Vec3ToCSV(gaze.leftEye.start) + ";" + *Vec3ToCSV(gaze.rightEye.start) + ";" + *Vec3ToCSV(gaze.gaze.start) + ";";
		}
		
		if (gaze.status == varjo_GazeStatus_Invalid || gaze.targetPos == FVector::ZeroVector)
		{
			dataString += "(Invalid);Invalid;Invalid;Invalid;Invalid;";
		}
		else
		{
			dataString += "(" + Vec3ToCSV(gaze.targetPos) + ");";
			dataString += FString::Printf(TEXT("%f;"), gaze.targetPos.Z);
			dataString += FString::Printf(TEXT("%f;"), RayToPointDistance(gaze.targetPos, gaze.leftEye.start, gaze.leftEye.dir));
			dataString += FString::Printf(TEXT("%f;"), RayToPointDistance(gaze.targetPos, gaze.rightEye.start, gaze.rightEye.dir));
			dataString += FString::Printf(TEXT("%f;"), RayToPointDistance(gaze.targetPos, gaze.gaze.start, gaze.gaze.dir));
		}

		dataString += LINE_TERMINATOR;
	}
	
	return FFileHelper::SaveStringToFile(dataString, *SavePath);
}

