// Copyright 6/4/2019 Varjo Technologies Oy. All Rights Reserved.

#include "VarjoGaze.h"
#include "DrawDebugHelpers.h"
#include "Engine/GameEngine.h"
#include "IVarjoHMDPlugin.h"
#include "IVarjoVREyeTracker.h"
#include "Gameframework/PlayerController.h"
#include "../../VarjoHMD/Private/VarjoHMD.h"

DEFINE_LOG_CATEGORY(LogVarjoGaze);

static FVarjoHMD* GetVarjoHMD()
{
	static FName SystemName(TEXT("VarjoHMD"));

	if (GEngine == nullptr)
	{
		return nullptr;
	}

	if (GEngine->XRSystem.IsValid() && (GEngine->XRSystem->GetSystemName() == SystemName))
	{
		return static_cast<FVarjoHMD*>(GEngine->XRSystem.Get());
	}
	UE_LOG(LogVarjoGaze, Warning, TEXT("Varjo HMD is not connected, eye tracking will be disabled."));
	return nullptr;
}

bool VarjoGaze::Init()
{
	check(IsInGameThread());
	if (m_initialized)
	{
		return false;
	}

	FlushRenderingCommands();
	m_varjoHMD = GetVarjoHMD();
	if (m_varjoHMD == nullptr)
	{
		return false;
	}

	m_session = m_varjoHMD->m_session;
	
	if (varjo_IsGazeAllowed(m_session) == false)
	{
		return false;
	}

	varjo_GazeInit(m_session);
	varjo_Error err = varjo_GetError(m_session);
	FMemory::Memzero(m_eyeTrackingData);

	if (err != varjo_NoError)
	{
		varjo_ClearError(m_session);
		return false;
	}
	m_focusDistance = 0;
	m_initialized = true;
	EyeTrackingStatus = EVarjoEyeTrackingStatus::NotCalibrated;
	return m_initialized;
}

void VarjoGaze::getRay(FVector& start, FVector& direction, float& focusDistance)
{
	check(IsInGameThread());
	start = m_start;
	direction = m_ray;
	focusDistance = m_focusDistance;
}

varjo_Gaze VarjoGaze::getGaze()
{
	if (m_initialized)
	{
		return varjo_Gaze();
	}

	varjo_Gaze result = varjo_GetGaze(m_session);
	if (varjo_GetError(m_session) != 0)
	{
		varjo_ClearError(m_session);
	}

	return result;
}

bool VarjoGaze::getGazeArray(TArray<varjo_Gaze>& gazeData)
{
	if (m_initialized == false)
	{
		return false;
	}

	FlushRenderingCommands();

	const int maxSize = 100;
	int lastSize = gazeData.AddUninitialized(maxSize);
	int size = varjo_GetGazeArray(m_session, gazeData.GetData() + lastSize, maxSize);
	gazeData.SetNum(lastSize + size);
	if (varjo_GetError(m_session) != 0) 
	{
		varjo_ClearError(m_session);
		return false;
	}
	return true;
}


FVarjoEyeTrackingRay VarjoGaze::VarjoRayConvert(const varjo_Ray& varjoRay)
{
	FVarjoEyeTrackingRay ray;
	ray.start = FVector(varjoRay.origin[0], varjoRay.origin[1], varjoRay.origin[2]);
	ray.dir = FVector(varjoRay.forward[0], varjoRay.forward[1], varjoRay.forward[2]);
	return std::move(ray);
}


FVarjoEyeTrackingData VarjoGaze::VarjoGazeDataConvert(const varjo_Gaze& gaze)
{
	FVarjoEyeTrackingData gazeDataBlueprint;
	gazeDataBlueprint.leftEye = VarjoRayConvert(gaze.leftEye);
	gazeDataBlueprint.rightEye = VarjoRayConvert(gaze.rightEye);
	gazeDataBlueprint.gaze = VarjoRayConvert(gaze.gaze);
	gazeDataBlueprint.targetId = -1;

	gazeDataBlueprint.frameNumber = (int)gaze.frameNumber;
	gazeDataBlueprint.focusDistance = (float)gaze.focusDistance;
	gazeDataBlueprint.captureTime = (int)gaze.captureTime;
	gazeDataBlueprint.rightPupilSize = (float)gaze.rightPupilSize;		
	gazeDataBlueprint.rightStatus = (int)gaze.rightStatus;
	gazeDataBlueprint.rawData = gaze;

	return std::move(gazeDataBlueprint);
}

bool VarjoGaze::getEyeTrackingData(TArray<FVarjoEyeTrackingData>& trackingData)
{
	TArray<varjo_Gaze> gazeData;
	if (getGazeArray(gazeData) == false)
	{
		return false;
	}

	const int maxSize = 100;
	trackingData.Reserve(trackingData.Num() + maxSize);

	for (auto& gaze : gazeData)
	{
		trackingData.Add(VarjoGazeDataConvert(gaze));
	}

	return true;
}

void VarjoGaze::requestGazeCalibration()
{
	varjo_RequestGazeCalibration(m_session);
	if (varjo_GetError(m_session) != 0)
	{
		varjo_ClearError(m_session);
	}
}


void VarjoGaze::drawDebug(UWorld* world)
{
	if (m_ray == FVector::ZeroVector)
	{
		return;
	}

	DrawDebugLine(world, m_start, m_start + m_ray * 10000.0f, FColor::Green, false, 5.0f, SDPG_Foreground, 1.0f);
}


void VarjoGaze::SetActivePlayerController(APlayerController* NewActivePlayerController)
{
	if (NewActivePlayerController && NewActivePlayerController->IsValidLowLevel() && NewActivePlayerController != ActivePlayerController)
	{
		ActivePlayerController = NewActivePlayerController;
	}
}

void VarjoGaze::Tick(float DeltaTime)
{
	check(IsInGameThread());

	if (m_initialized == false)
	{
		return;
	}

	UpdateStatus();
	varjo_Gaze gaze = varjo_GetGaze(m_session);

	if (varjo_GetError(m_session) != 0)
	{
		varjo_ClearError(m_session);
	}

	m_invViewQuat = m_varjoHMD->InvViewQuat;
	m_invViewOrigin = m_varjoHMD->InvViewOrigin;

	if (gaze.status == varjo_GazeStatus_Valid)
	{
		FVector dir = FVector(gaze.gaze.forward[0], gaze.gaze.forward[1], gaze.gaze.forward[2]);
		m_ray = m_invViewQuat * dir;
		m_eyeTrackingData = VarjoGazeDataConvert(gaze);
		m_focusDistance = gaze.focusDistance * m_varjoHMD->GetWorldToMetersScale();
		m_start = m_invViewOrigin + FVector(-gaze.gaze.origin[2], gaze.gaze.origin[0], gaze.gaze.origin[1]) * m_varjoHMD->GetWorldToMetersScale();
	}
	else
	{
		m_start = FVector::ZeroVector;
		m_ray = FVector::ZeroVector;
		FMemory::Memzero(m_eyeTrackingData);
		m_eyeTrackingData.targetId = -1;
	}

	if (m_debugViewEnabled && GEngine != nullptr && GEngine->GetWorld() != nullptr)
	{
		drawDebug(GEngine->GetWorld());
	}
}

const FVarjoEyeTrackingData& VarjoGaze::GetEyeTrackingData()
{
	return m_eyeTrackingData;
}

const FVector VarjoGaze::GetFixationPoint()
{
	return m_start + m_ray*m_eyeTrackingData.focusDistance;
}

EVarjoEyeTrackingStatus VarjoGaze::GetEyeTrackingStatus()
{
	UpdateStatus();
	return EyeTrackingStatus;
}

EVarjoEyeTrackingCalibrationStatus VarjoGaze::GetCalibrationStatus()
{
	UpdateStatus();
	return EyeCalibrationStatus;
}

void VarjoGaze::UpdateStatus()
{
	if (m_initialized == false)
	{
		if (Init() == false)
		{
			return;
		}
	}

	if (m_eyeTrackingData.rawData.status == varjo_GazeStatus_Valid)
	{
		EyeTrackingStatus = EVarjoEyeTrackingStatus::Calibrated;
		EyeCalibrationStatus = m_eyeTrackingData.rawData.leftStatus > 0 && m_eyeTrackingData.rawData.rightStatus > 0 ? EVarjoEyeTrackingCalibrationStatus::Good : EVarjoEyeTrackingCalibrationStatus::Bad;
	}
	else
	{
		EyeTrackingStatus = EVarjoEyeTrackingStatus::NotCalibrated;
		EyeCalibrationStatus = EVarjoEyeTrackingCalibrationStatus::None;
	}
}
