// Copyright 6/4/2019 Varjo Technologies Oy. All Rights Reserved.

#pragma once

#include "IVarjoVRControllerPlugin.h"
#include "IVarjoHMDPlugin.h"
#include "IInputDevice.h"
#include "IHapticDevice.h"
#include "XRMotionControllerBase.h"
#include "Engine/Engine.h"
#include "EngineGlobals.h"
#include "GenericPlatform/IInputInterface.h"
#include "../../VarjoHMD/Private/VarjoHMD.h"
#include "VarjoVRControllerLibrary.h"

#define LOCTEXT_NAMESPACE "VarjoVRController"

#if PLATFORM_WINDOWS
#include "Windows/WindowsHWrapper.h"
#endif

#include "openvr.h"

DEFINE_LOG_CATEGORY_STATIC(LogVarjoVRController, Log, All);

/** Total number of controllers in a set */
#define CONTROLLERS_PER_PLAYER	2

#define MAX_TRACKED_DEVICES (int32)EControllerHand::Special_9 - (int32)EControllerHand::Left + 1

/** Player that generic trackers will be assigned to */
#define GENERIC_TRACKER_PLAYER_NUM 0

/** Controller axis mappings. @todo varjohmd: should enumerate rather than hard code */
#define TOUCHPAD_AXIS					0
#define TRIGGER_AXIS					1
#define KNUCKLES_TOTAL_HAND_GRIP_AXIS	2
#define KNUCKLES_UPPER_HAND_GRIP_AXIS	3
#define KNUCKLES_LOWER_HAND_GRIP_AXIS	4
#define DOT_45DEG		0.7071f

//
// Gamepad thresholds
//
#define TOUCHPAD_DEADZONE  0.0f

// Controls whether or not we need to swap the input routing for the hands, for debugging
static TAutoConsoleVariable<int32> CVarSwapHands(
	TEXT("vr.SwapMotionControllerInput"),
	0,
	TEXT("This command allows you to swap the button / axis input handedness for the input controller, for debugging purposes.\n")
	TEXT(" 0: don't swap (default)\n")
	TEXT(" 1: swap left and right buttons"),
	ECVF_Cheat);

namespace VarjoVRControllerKeyNames
{
	const FGamepadKeyNames::Type Touch0("Steam_Touch_0");
	const FGamepadKeyNames::Type Touch1("Steam_Touch_1");
	const FGamepadKeyNames::Type GenericGrip("Steam_Generic_Grip");
	const FGamepadKeyNames::Type GenericTrigger("Steam_Generic_Trigger");
	const FGamepadKeyNames::Type GenericTouchpad("Steam_Generic_Touchpad");
	const FGamepadKeyNames::Type GenericMenu("Steam_Generic_Menu");
	const FGamepadKeyNames::Type GenericSystem("Steam_Generic_System");

	const FGamepadKeyNames::Type VarjoVR_Knuckles_Left_HandGrip("VarjoVR_Knuckles_Left_HandGrip");
	const FGamepadKeyNames::Type VarjoVR_Knuckles_Left_IndexGrip("VarjoVR_Knuckles_Left_IndexGrip");
	const FGamepadKeyNames::Type VarjoVR_Knuckles_Left_MiddleGrip("VarjoVR_Knuckles_Left_MiddleGrip");
	const FGamepadKeyNames::Type VarjoVR_Knuckles_Left_RingGrip("VarjoVR_Knuckles_Left_RingGrip");
	const FGamepadKeyNames::Type VarjoVR_Knuckles_Left_PinkyGrip("VarjoVR_Knuckles_Left_PinkyGrip");

	const FGamepadKeyNames::Type VarjoVR_Knuckles_Right_HandGrip("VarjoVR_Knuckles_Right_HandGrip");
	const FGamepadKeyNames::Type VarjoVR_Knuckles_Right_IndexGrip("VarjoVR_Knuckles_Right_IndexGrip");
	const FGamepadKeyNames::Type VarjoVR_Knuckles_Right_MiddleGrip("VarjoVR_Knuckles_Right_MiddleGrip");
	const FGamepadKeyNames::Type VarjoVR_Knuckles_Right_RingGrip("VarjoVR_Knuckles_Right_RingGrip");
	const FGamepadKeyNames::Type VarjoVR_Knuckles_Right_PinkyGrip("VarjoVR_Knuckles_Right_PinkyGrip");
}

namespace VarjoVRControllerKeys
{
	const FKey VarjoVR_Knuckles_Left_HandGrip("VarjoVR_Knuckles_Left_HandGrip");
	const FKey VarjoVR_Knuckles_Left_IndexGrip("VarjoVR_Knuckles_Left_IndexGrip");
	const FKey VarjoVR_Knuckles_Left_MiddleGrip("VarjoVR_Knuckles_Left_MiddleGrip");
	const FKey VarjoVR_Knuckles_Left_RingGrip("VarjoVR_Knuckles_Left_RingGrip");
	const FKey VarjoVR_Knuckles_Left_PinkyGrip("VarjoVR_Knuckles_Left_PinkyGrip");

	const FKey VarjoVR_Knuckles_Right_HandGrip("VarjoVR_Knuckles_Right_HandGrip");
	const FKey VarjoVR_Knuckles_Right_IndexGrip("VarjoVR_Knuckles_Right_IndexGrip");
	const FKey VarjoVR_Knuckles_Right_MiddleGrip("VarjoVR_Knuckles_Right_MiddleGrip");
	const FKey VarjoVR_Knuckles_Right_RingGrip("VarjoVR_Knuckles_Right_RingGrip");
	const FKey VarjoVR_Knuckles_Right_PinkyGrip("VarjoVR_Knuckles_Right_PinkyGrip");
}

class FVarjoVRController : public IInputDevice, public FXRMotionControllerBase, public IHapticDevice
{
public:
	FVarjoHMD * GetVarjoHMD() const;

	/** The maximum number of Unreal controllers.  Each Unreal controller represents a pair of motion controller devices */
	static const int32 MaxUnrealControllers = MAX_VARJOVR_CONTROLLER_PAIRS;

	/** Total number of motion controllers we'll support */
	static const int32 MaxControllers = MaxUnrealControllers * CONTROLLERS_PER_PLAYER;

	/** The maximum number of Special hand designations available to use for generic trackers
	*  Casting enums directly, so if the input model changes, this won't silently be invalid
	*/
	static const int32 MaxSpecialDesignations = (int32)EControllerHand::Special_9 - (int32)EControllerHand::Special_1 + 1;

	/**
	* Buttons on the VarjoVR controller
	*/
	struct EVarjoVRControllerButton
	{
		enum Type
		{
			System,
			ApplicationMenu,
			TouchPadPress,
			TouchPadTouch,
			TriggerPress,
			Grip,
			TouchPadUp,
			TouchPadDown,
			TouchPadLeft,
			TouchPadRight,

			/** Max number of controller buttons.  Must be < 256 */
			TotalButtonCount
		};
	};

	FVarjoVRController(const TSharedRef< FGenericApplicationMessageHandler >& InMessageHandler);
	virtual ~FVarjoVRController();

	virtual void Tick(float DeltaTime) override;
	virtual void SendControllerEvents() override;
	void SetTouchDPadMapping(EVarjoVRTouchDPadMapping NewMapping);
	int32 UnrealControllerIdToControllerIndex(const int32 UnrealControllerId, const EControllerHand Hand) const;
	void SetChannelValue(int32 UnrealControllerId, FForceFeedbackChannelType ChannelType, float Value) override;
	void SetChannelValues(int32 UnrealControllerId, const FForceFeedbackValues& Values) override;
	virtual IHapticDevice* GetHapticDevice() override;
	virtual void SetHapticFeedbackValues(int32 UnrealControllerId, int32 Hand, const FHapticFeedbackValues& Values) override;
	virtual void GetHapticFrequencyRange(float& MinFrequency, float& MaxFrequency) const override;
	virtual float GetHapticAmplitudeScale() const override;
	void UpdateVibration(const int32 ControllerIndex);
	virtual void SetMessageHandler(const TSharedRef< FGenericApplicationMessageHandler >& InMessageHandler) override;
	virtual bool Exec(UWorld* InWorld, const TCHAR* Cmd, FOutputDevice& Ar) override;
	virtual FName GetMotionControllerDeviceTypeName() const override;
	virtual bool GetControllerOrientationAndPosition(const int32 ControllerIndex, const EControllerHand DeviceHand, FRotator& OutOrientation, FVector& OutPosition, float WorldToMetersScale) const;
	bool IsControllerConnected(const int32 ControllerIndex, const EControllerHand DeviceHand) const;
	float GetControllerBatteryLevel(const int32 ControllerIndex, const EControllerHand DeviceHand) const;
	virtual ETrackingStatus GetControllerTrackingStatus(const int32 ControllerIndex, const EControllerHand DeviceHand) const;
	virtual bool IsGamepadAttached() const override;

	static FName DeviceTypeName;
	static EVarjoVRTouchDPadMapping DefaultDPadMapping;
private:

	inline vr::IVRSystem* GetVRSystem();

	void RegisterDeviceChanges(vr::IVRSystem* VRSystem);
	bool RegisterController(uint32 DeviceIndex, vr::IVRSystem* VRSystem);
	void DetectHandednessSwap(vr::IVRSystem* VRSystem);
	bool RegisterTracker(uint32 DeviceIndex);
	void UnregisterController(uint32 DeviceIndex);
	void UnregisterTracker(uint32 DeviceIndex);
	void UnregisterDevice(uint32 DeviceIndex);

	struct FControllerState
	{
		/** Which hand this controller is representing */
		EControllerHand Hand;

		/** If packet num matches that on your prior call, then the controller state hasn't been changed since
		* your last call and there is no need to process it. */
		uint32 PacketNum;

		/** touchpad analog values */
		float TouchPadXAnalog;
		float TouchPadYAnalog;

		/** trigger analog value */
		float TriggerAnalog;

		/** Knuckles Controller Axes */
		float HandGripAnalog;
		float IndexGripAnalog;
		float MiddleGripAnalog;
		float RingGripAnalog;
		float PinkyGripAnalog;

		/** Last frame's button states, so we only send events on edges */
		bool ButtonStates[EVarjoVRControllerButton::TotalButtonCount];

		/** Next time a repeat event should be generated for each button */
		double NextRepeatTime[EVarjoVRControllerButton::TotalButtonCount];

		/** Value for force feedback on this controller hand */
		float ForceFeedbackValue;
	};

	/** Mappings between tracked devices and 0 indexed controllers */
	int32 NumControllersMapped;
	int32 NumTrackersMapped;
	int32 DeviceToControllerMap[vr::k_unMaxTrackedDeviceCount];
	int32 UnrealControllerIdAndHandToDeviceIdMap[MaxUnrealControllers][vr::k_unMaxTrackedDeviceCount];
	int32 UnrealControllerHandUsageCount[CONTROLLERS_PER_PLAYER];

	/** Controller states */
	FControllerState ControllerStates[MaxControllers];

	/** Delay before sending a repeat message after a button was first pressed */
	float InitialButtonRepeatDelay;

	/** Delay before sending a repeat message after a button has been pressed for a while */
	float ButtonRepeatDelay;

	/** Mapping of controller buttons */
	FGamepadKeyNames::Type Buttons[vr::k_unMaxTrackedDeviceCount][EVarjoVRControllerButton::TotalButtonCount];

	/** weak pointer to the IVRSystem owned by the HMD module */
	TWeakPtr<vr::IVRSystem> HMDVRSystem;

	/** handler to send all messages to */
	TSharedRef<FGenericApplicationMessageHandler> MessageHandler;

	/** the VarjoHMD plugin module */
	IVarjoHMDPlugin* VarjoHMDPlugin;
};

#undef LOCTEXT_NAMESPACE //"VarjoVRController"
