// Copyright 2019 Varjo Technologies Oy. All rights reserved.

#ifndef VARJO_MR_H
#define VARJO_MR_H

#include "Varjo_types.h"
#include "Varjo_types_mr.h"

/**
 * Varjo Mixed Reality API (EXPERIMENTAL)
 *
 * API is not yet stable, so beware that any functionality in this file might change / get removed.
 */

#if defined __cplusplus
extern "C" {
#endif

/**
 * Start video see-through rendering on the headset displays.
 *
 * Can be called from multiple applications. Rendering will continue until all applications have stopped
 * the rendering with varjo_MRStopRender (or disconnected).
 *
 * Video is rendered as a background layer. i.e. Any application rendering is done on top of the video.
 *
 * @param session Varjo session handle.
 */
VARJO_API void varjo_MRStartRender(struct varjo_Session* session);

/**
 * Stop video see-through rendering.
 *
 * @param session Varjo session handle.
 */
VARJO_API void varjo_MRStopRender(struct varjo_Session* session);

/**
 * Start camera frame streaming into the provided callback.
 *
 * Multiple applications can listen to the camera stream simultaneously.
 *
 * Camera frames are not queued. If the previous callback has not been
 * finished when new frames arrive, the new frames are dropped.
 *
 * Note that the callback is not quaranteed to be always called from the same thread.
 *
 * @param session Varjo session handle.
 * @param flags Flags to control the camera feed.
 * @param callback Callback which receives the new frames.
 */
VARJO_API void varjo_MRStartFrameStream(struct varjo_Session* session, varjo_MRCameraFlag flags, varjo_FrameListener callback);

/**
 * Stop camera frame streaming.
 *
 * There might be calls into the callback while executing this function, but after the function
 * has returned the callback is guaranteed not to be called anymore.
 *
 * @param session Varjo session handle.
 */
VARJO_API void varjo_MRStopFrameStream(struct varjo_Session* session);

/**
 * Try to lock the camera configuration.
 *
 * This call needs to succeed before any varjo_MRCameraSet*-calls.
 * varjo_MRCameraUnlockConfig() should be called after the configuration has been changed.
 *
 * The configuration can be locked by one application at a time. Locking fails,
 * if another client has the configuration locked already.
 *
 * Note that if an application wants to prevent others from changing the camera settings,
 * the lock can be held for a long period of time.
 *
 * @param session Varjo session handle.
 * @return varjo_True if locked successfully, otherwise varjo_False.
 */
VARJO_API varjo_Bool varjo_MRCameraTryLockConfig(struct varjo_Session* session);

/**
 * Unlock previously locked camera configuration.
 *
 * Note that after unlocking the configuration other applications are free to modify the settings.
 *
 * @param session Varjo session handle.
 */
VARJO_API void varjo_MRCameraUnlockConfig(struct varjo_Session* session);

/**
 * Query if a certain mode is supported.
 *
 * Should be called before Set*-functions to determine if the camera supports the desired mode.
 *
 * Modes can be queried once and cached as they are not going to change during application runtime.
 *
 * @param session Varjo session handle.
 * @param mode Mode to check.
 * @return varjo_True if the mode is supported, otherwise varjo_False.
 */
VARJO_API varjo_Bool varjo_MRCameraSupportsWhiteBalanceMode(struct varjo_Session* session, varjo_MRCamera_WhiteBalanceMode mode);
VARJO_API varjo_Bool varjo_MRCameraSupportsISOMode(struct varjo_Session* session, varjo_MRCamera_ISOMode mode);
VARJO_API varjo_Bool varjo_MRCameraSupportsExposureMode(struct varjo_Session* session, varjo_MRCamera_ExposureMode mode);
VARJO_API varjo_Bool varjo_MRCameraSupportsEVCompensation(struct varjo_Session* session, varjo_MRCamera_EVCompensation mode);

/**
 * Set camera mode.
 *
 * varjo_MRCameraTryLockConfig() has to be called successfully before calling the setters.
 *
 * @param session Varjo session handle.
 * @param mode New camera mode.
 */
VARJO_API void varjo_MRCameraSetWhiteBalanceMode(struct varjo_Session* session, varjo_MRCamera_WhiteBalanceMode mode);
VARJO_API void varjo_MRCameraSetISOMode(struct varjo_Session* session, varjo_MRCamera_ISOMode mode);
VARJO_API void varjo_MRCameraSetExposureMode(struct varjo_Session* session, varjo_MRCamera_ExposureMode mode);
VARJO_API void varjo_MRCameraSetEVCompensation(struct varjo_Session* session, varjo_MRCamera_EVCompensation mode);

/**
 * Get current camera modes.
 *
 * Modes can be queried once and cached as they are not going to change during application runtime.
 *
 * @param session Varjo session handle.
 * @return Current mode. varjo_MRCameraMode_Invalid in case of an error.
 */
VARJO_API varjo_MRCamera_ExposureMode varjo_MRCameraGetExposureMode(struct varjo_Session* session);
VARJO_API varjo_MRCamera_ISOMode varjo_MRCameraGetISOMode(struct varjo_Session* session);
VARJO_API varjo_MRCamera_WhiteBalanceMode varjo_MRCameraGetWhiteBalanceMode(struct varjo_Session* session);
VARJO_API varjo_MRCamera_EVCompensation varjo_MRCameraGetEVCompensation(struct varjo_Session* session);

#if defined __cplusplus
}
#endif

#endif  // VARJO_MR_H
