// Copyright 2019 Varjo Technologies Oy. All rights reserved.

#ifndef VARJO_TYPES_MR_H
#define VARJO_TYPES_MR_H

#include "Varjo_types.h"

/**
 * Varjo Mixed Reality API - types (EXPERIMENTAL)
 *
 * API is not yet stable, so beware that any functionality in this file might change / get removed.
 */

#if defined __cplusplus
extern "C" {
#endif

/**
 * Mixed reality errors (700 - 799)
 */
static const varjo_Error varjo_Error_MRConnectionFailed = 700;
static const varjo_Error varjo_Error_MRCameraStartFailed = 701;
static const varjo_Error varjo_Error_MRStreamingStartFailed = 702;
static const varjo_Error varjo_Error_MRInvalidFrameCallback = 703;
static const varjo_Error varjo_Error_MRCameraConfigNotLocked = 704;
static const varjo_Error varjo_Error_MRCameraConfigInvalidMode = 705;
static const varjo_Error varjo_Error_MROperationFailed = 706;

/**
 * Property key to check for Mixed Reality support.
 */
static const varjo_PropertyKey varjo_PropertyKey_MRAvailable = 0xD000;

/**
 * Invalid camera mode (white balance, ISO, exposure or ev compensation).
 */
static const int64_t varjo_MRCameraMode_Invalid = -1;

/**
 * White balance modes.
 */
typedef int64_t varjo_MRCamera_WhiteBalanceMode;
static const varjo_MRCamera_WhiteBalanceMode varjo_MRCamera_WhiteBalanceMode_Auto = 0;
static const varjo_MRCamera_WhiteBalanceMode varjo_MRCamera_WhiteBalanceMode_12000K = 1;
static const varjo_MRCamera_WhiteBalanceMode varjo_MRCamera_WhiteBalanceMode_8000K = 2;
static const varjo_MRCamera_WhiteBalanceMode varjo_MRCamera_WhiteBalanceMode_5000K = 3;
static const varjo_MRCamera_WhiteBalanceMode varjo_MRCamera_WhiteBalanceMode_3000K = 4;
static const varjo_MRCamera_WhiteBalanceMode varjo_MRCamera_WhiteBalanceMode_2000K = 5;

/**
 * ISO modes.
 */
typedef int64_t varjo_MRCamera_ISOMode;
static const varjo_MRCamera_ISOMode varjo_MRCamera_ISOMode_Auto = 0;
static const varjo_MRCamera_ISOMode varjo_MRCamera_ISOMode_ISO100 = 1;
static const varjo_MRCamera_ISOMode varjo_MRCamera_ISOMode_ISO200 = 2;
static const varjo_MRCamera_ISOMode varjo_MRCamera_ISOMode_ISO400 = 3;
static const varjo_MRCamera_ISOMode varjo_MRCamera_ISOMode_ISO800 = 4;
static const varjo_MRCamera_ISOMode varjo_MRCamera_ISOMode_ISO1600 = 5;
static const varjo_MRCamera_ISOMode varjo_MRCamera_ISOMode_ISO3200 = 6;
static const varjo_MRCamera_ISOMode varjo_MRCamera_ISOMode_ISO6400 = 7;

/**
 * Exposure modes.
 */
typedef int64_t varjo_MRCamera_ExposureMode;
static const varjo_MRCamera_ExposureMode varjo_MRCamera_ExposureMode_Auto = 0;
static const varjo_MRCamera_ExposureMode varjo_MRCamera_ExposureMode_1_90_sec = 1;
static const varjo_MRCamera_ExposureMode varjo_MRCamera_ExposureMode_1_125_sec = 2;
static const varjo_MRCamera_ExposureMode varjo_MRCamera_ExposureMode_1_250_sec = 3;
static const varjo_MRCamera_ExposureMode varjo_MRCamera_ExposureMode_1_500_sec = 4;
static const varjo_MRCamera_ExposureMode varjo_MRCamera_ExposureMode_1_1000_sec = 5;
static const varjo_MRCamera_ExposureMode varjo_MRCamera_ExposureMode_1_2000_sec = 6;
static const varjo_MRCamera_ExposureMode varjo_MRCamera_ExposureMode_1_4000_sec = 7;
static const varjo_MRCamera_ExposureMode varjo_MRCamera_ExposureMode_1_8000_sec = 8;

/**
 * Exposure compensation values.
 */
typedef int64_t varjo_MRCamera_EVCompensation;
static const varjo_MRCamera_EVCompensation varjo_MRCamera_EVCompensation_0 = 0;
static const varjo_MRCamera_EVCompensation varjo_MRCamera_EVCompensation_plus_3_0 = 1;
static const varjo_MRCamera_EVCompensation varjo_MRCamera_EVCompensation_plus_2_0 = 2;
static const varjo_MRCamera_EVCompensation varjo_MRCamera_EVCompensation_plus_1_0 = 3;
static const varjo_MRCamera_EVCompensation varjo_MRCamera_EVCompensation_minus_1_0 = 4;
static const varjo_MRCamera_EVCompensation varjo_MRCamera_EVCompensation_minus_2_0 = 5;
static const varjo_MRCamera_EVCompensation varjo_MRCamera_EVCompensation_minus_3_0 = 6;

/**
 * Control flags for camera stream.
 */
typedef uint32_t varjo_MRCameraFlag;
static const varjo_MRCameraFlag varjo_MRCameraFlag_ExcludeRight = 0x1;  //!< Right channel camera frame will not be included in camera callback.
static const varjo_MRCameraFlag varjo_MRCameraFlag_ExcludeLeft = 0x2;   //!< Left channel camera frame will not be included in camera callback.

/**
 * Raw camera frame data.
 */
struct varjo_MRCameraFrameData {
    varjo_TextureFormat format;       //!< Timestamp at end of exposure.
    void* buffer;                     //!< Buffer containing the pixel data.
    uint32_t bufferByteSize;          //!< Buffer size in bytes.
    uint32_t rowStride;               //!< Buffer horizontal stride in bytes.
    uint32_t width;                   //!< Image width.
    uint32_t height;                  //!< Image height.
    float principalPointX;            //!< Camera principal point X.
    float principalPointY;            //!< Camera principal point Y.
    float focalLengthX;               //!< Camera focal length X.
    float focalLengthY;               //!< Camera focal length Y.
    float distortionCoefficients[6];  //!< OpenCV omnidir form: 2 radial, skew, xi, 2 tangential
};

/**
 * Frame stream callback data.
 */
struct varjo_MRCameraFrame {
    varjo_Nanoseconds timestamp;            //!< Timestamp at end of exposure.
    struct varjo_MRCameraFrameData* left;   //!< Frame data for the left camera.
    struct varjo_MRCameraFrameData* right;  //!< Frame data for the right camera.
    float evValue;                          //!< EV value at ISO100.
    float exposureTime;                     //!< Exposure time in seconds.
    float whiteBalanceR;                    //!< White balance red component.
    float whiteBalanceG;                    //!< White balance green component.
    float whiteBalanceB;                    //!< White balance blue component.
    int32_t reserved;                       //!< Unused.
};

/**
 * Camera frame stream callback function.
 */
typedef void (*varjo_FrameListener)(const struct varjo_MRCameraFrame*);

#if defined __cplusplus
}
#endif

#endif  // VARJO_TYPES_MR_H
